
"""GNUTLS connection support"""

__all__ = ['X509Credentials', 'TLSContext', 'ClientSession', 'ServerSession', 'ServerSessionFactory']

from time import time
from socket import SHUT_RDWR as SOCKET_SHUT_RDWR

from _ctypes import PyObj_FromPtr
from ctypes import *

from gnutls.validators import *
from gnutls.constants import *
from gnutls.crypto import *
from gnutls.errors import *

from gnutls.library.constants import GNUTLS_SERVER, GNUTLS_CLIENT, GNUTLS_CRT_X509
from gnutls.library.constants import GNUTLS_CERT_INVALID, GNUTLS_CERT_REVOKED, GNUTLS_CERT_INSECURE_ALGORITHM
from gnutls.library.constants import GNUTLS_CERT_SIGNER_NOT_FOUND, GNUTLS_CERT_SIGNER_NOT_CA
from gnutls.library.constants import GNUTLS_AL_FATAL, GNUTLS_A_BAD_CERTIFICATE
from gnutls.library.constants import GNUTLS_A_UNKNOWN_CA, GNUTLS_A_INSUFFICIENT_SECURITY
from gnutls.library.constants import GNUTLS_A_CERTIFICATE_EXPIRED, GNUTLS_A_CERTIFICATE_REVOKED
from gnutls.library.constants import GNUTLS_NAME_DNS
from gnutls.library.types     import gnutls_certificate_credentials_t, gnutls_session_t, gnutls_x509_crt_t
from gnutls.library.types     import gnutls_certificate_retrieve_function
from gnutls.library.types     import gnutls_priority_t
from gnutls.library.functions import *


@gnutls_certificate_retrieve_function
def _retrieve_certificate(c_session, req_ca_dn, nreqs, pk_algos, pk_algos_length, retr_st):
    session = PyObj_FromPtr(gnutls_session_get_ptr(c_session))
    identity = session.credentials.select_server_identity(session)
    retr_st.contents.deinit_all = 0
    if identity is None:
        retr_st.contents.ncerts = 0
    else:
        retr_st.contents.ncerts = 1
        retr_st.contents.cert_type = GNUTLS_CRT_X509
        retr_st.contents.cert.x509.contents = identity.cert._c_object
        retr_st.contents.key.x509 = identity.key._c_object
    return 0


class _ServerNameIdentities(dict):
    """Used internally by X509Credentials to map server names to X509 identities for the server name extension"""
    def __init__(self, identities):
        dict.__init__(self)
        for identity in identities:
            self.add(identity)
    def add(self, identity):
        for name in identity.cert.alternative_names.dns:
            self[name.lower()] = identity
        for ip in identity.cert.alternative_names.ip:
            self[ip] = identity
        subject = identity.cert.subject
        if subject.CN is not None:
            self[subject.CN.lower()] = identity
    def get(self, server_name, default=None):
        server_name = server_name.lower()
        if server_name in self:
            return self[server_name]
        for name in (n for n in self if n.startswith('*.')):
            suffix = name[1:]
            if server_name.endswith(suffix) and '.' not in server_name[:-len(suffix)]:
                return self[name]
        return default


class X509Credentials(object):
    def __new__(cls, *args, **kwargs):
        c_object = gnutls_certificate_credentials_t()
        gnutls_certificate_allocate_credentials(byref(c_object))
        instance = object.__new__(cls)
        instance.__deinit = gnutls_certificate_free_credentials
        instance._c_object = c_object
        return instance

    @method_args((X509Certificate, none), (X509PrivateKey, none), list_of(X509Certificate), list_of(X509CRL), list_of(X509Identity))
    def __init__(self, cert=None, key=None, trusted=[], crl_list=[], identities=[]):
        """Credentials contain a X509 certificate, a private key, a list of trusted CAs and a list of CRLs (all optional).
        An optional list of additional X509 identities can be specified for applications that need more that one identity"""
        if cert and key:
            gnutls_certificate_set_x509_key(self._c_object, byref(cert._c_object), 1, key._c_object)
        elif (cert, key) != (None, None):
            raise ValueError("Specify neither or both the certificate and private key")
        gnutls_certificate_set_retrieve_function(self._c_object, _retrieve_certificate)
        self._max_depth = 5
        self._max_bits  = 8200
        self._type = CRED_CERTIFICATE
        self._cert = cert
        self._key = key
        self._identities = tuple(identities)
        self._trusted = ()
        self.add_trusted(trusted)
        self.crl_list = crl_list
        self.server_name_identities = _ServerNameIdentities(identities)
        if cert and key:
            self.server_name_identities.add(X509Identity(cert, key))

    def __del__(self):
        self.__deinit(self._c_object)

    # Methods to alter the credentials at runtime

    @method_args(list_of(X509Certificate))
    def add_trusted(self, trusted):
        size = len(trusted)
        if size > 0:
            ca_list = (gnutls_x509_crt_t * size)(*[cert._c_object for cert in trusted])
            gnutls_certificate_set_x509_trust(self._c_object, cast(byref(ca_list), POINTER(gnutls_x509_crt_t)), size)
            self._trusted = self._trusted + tuple(trusted)

    # Properties

    @property
    def cert(self):
        return self._cert

    @property
    def key(self):
        return self._key

    @property
    def identities(self):
        return self._identities

    @property
    def trusted(self):
        return self._trusted

    def _get_crl_list(self):
        return self._crl_list
    @method_args(list_of(X509CRL)) 
    def _set_crl_list(self, crl_list):
        self._crl_list = tuple(crl_list)
    crl_list = property(_get_crl_list, _set_crl_list)
    del _get_crl_list, _set_crl_list

    def _get_max_verify_length(self):
        return self._max_depth
    @method_args(int) 
    def _set_max_verify_length(self, max_depth):
        gnutls_certificate_set_verify_limits(self._c_object, self._max_bits, max_depth)
        self._max_depth = max_depth
    max_verify_length = property(_get_max_verify_length, _set_max_verify_length)
    del _get_max_verify_length, _set_max_verify_length

    def _get_max_verify_bits(self):
        return self._max_bits
    @method_args(int) 
    def _set_max_verify_bits(self, max_bits):
        gnutls_certificate_set_verify_limits(self._c_object, max_bits, self._max_depth)
        self._max_bits = max_bits
    max_verify_bits = property(_get_max_verify_bits, _set_max_verify_bits)
    del _get_max_verify_bits, _set_max_verify_bits

    # Methods to select and validate certificates

    def check_certificate(self, cert, cert_name='certificate'):
        """Verify activation, expiration and revocation for the given certificate"""
        now = time()
        if cert.activation_time > now:
            raise CertificateExpiredError("%s is not yet activated" % cert_name)
        if cert.expiration_time < now:
            raise CertificateExpiredError("%s has expired" % cert_name)
        for crl in self.crl_list:
            crl.check_revocation(cert, cert_name=cert_name)

    def select_server_identity(self, session):
        """Select which identity the server will use for a given session. The default selection algorithm uses
        the server name extension. A subclass can overwrite it if a different selection algorithm is desired."""
        server_name = session.server_name
        if server_name is not None:
            return self.server_name_identities.get(server_name)
        elif self.cert and self.key:
            return self ## since we have the cert and key attributes we can behave like a X509Identity
        else:
            return None


class TLSContext(object):
    def __init__(self, credentials, session_parameters=None):
        self.credentials = credentials
        self.session_parameters = session_parameters

    @property
    def session_parameters(self):
        return self.__dict__.get('session_parameters')

    @session_parameters.setter
    def session_parameters(self, value):
        priority = gnutls_priority_t()
        try:
            gnutls_priority_init(byref(priority), value, None)
        except GNUTLSError:
            raise ValueError("invalid session parameters: %s" % value)
        else:
            gnutls_priority_deinit(priority)
        self.__dict__['session_parameters'] = value


class Session(object):
    """Abstract class representing a TLS session created from a TCP socket
       and a Credentials object."""

    session_type = None ## placeholder for GNUTLS_SERVER or GNUTLS_CLIENT as defined by subclass

    def __new__(cls, *args, **kwargs):
        if cls is Session:
            raise RuntimeError("Session cannot be instantiated directly")
        instance = object.__new__(cls)
        instance.__deinit = gnutls_deinit
        instance._c_object = gnutls_session_t()
        return instance

    def __init__(self, socket, context):
        gnutls_init(byref(self._c_object), self.session_type)
        ## Store a pointer to self on the C session
        gnutls_session_set_ptr(self._c_object, id(self))
        gnutls_set_default_priority(self._c_object)
        gnutls_priority_set_direct(self._c_object, context.session_parameters, None)
        gnutls_transport_set_ptr(self._c_object, socket.fileno())
        gnutls_handshake_set_private_extensions(self._c_object, 1)
        self.socket = socket
        self.credentials = context.credentials

    def __del__(self):
        self.__deinit(self._c_object)

    def __getattr__(self, name):
        ## Generic wrapper for the underlying socket methods and attributes.
        return getattr(self.socket, name)

    # Session properties

    def _get_credentials(self):
        return self._credentials
    @method_args(X509Credentials)
    def _set_credentials(self, credentials):
        ## Release all credentials, otherwise gnutls will only release an existing credential of
        ## the same type as the one being set and we can end up with multiple credentials in C.
        gnutls_credentials_clear(self._c_object)
        gnutls_credentials_set(self._c_object, credentials._type, cast(credentials._c_object, c_void_p))
        self._credentials = credentials
    credentials = property(_get_credentials, _set_credentials)
    del _get_credentials, _set_credentials

    @property
    def protocol(self):
        return gnutls_protocol_get_name(gnutls_protocol_get_version(self._c_object))

    @property
    def kx_algorithm(self):
        return gnutls_kx_get_name(gnutls_kx_get(self._c_object))

    @property
    def cipher(self):
        return gnutls_cipher_get_name(gnutls_cipher_get(self._c_object))

    @property
    def mac_algorithm(self):
        return gnutls_mac_get_name(gnutls_mac_get(self._c_object))

    @property
    def compression(self):
        return gnutls_compression_get_name(gnutls_compression_get(self._c_object))

    @property
    def peer_certificate(self):
        if gnutls_certificate_type_get(self._c_object) != GNUTLS_CRT_X509:
            return None
        list_size = c_uint()
        cert_list = gnutls_certificate_get_peers(self._c_object, byref(list_size))
        if list_size.value == 0:
            return None
        cert = cert_list[0]
        return X509Certificate(string_at(cert.data, cert.size), X509_FMT_DER)

    # Status checking after an operation was interrupted (these properties are
    # only useful to check after an operation was interrupted, otherwise their
    # value is meaningless).

    @property
    def interrupted_while_writing(self):
        """True if an operation was interrupted while writing"""
        return gnutls_record_get_direction(self._c_object)==1

    @property
    def interrupted_while_reading(self):
        """True if an operation was interrupted while reading"""
        return gnutls_record_get_direction(self._c_object)==0

    # Session methods

    def handshake(self):
        gnutls_handshake(self._c_object)

    def send(self, data):
        data = str(data)
        if not data:
            return 0
        return gnutls_record_send(self._c_object, data, len(data))

    def sendall(self, data):
        size = len(data)
        while size > 0:
            sent = self.send(data[-size:])
            size -= sent

    def recv(self, limit):
        data = create_string_buffer(limit)
        size = gnutls_record_recv(self._c_object, data, limit)
        return data[:size]

    def send_alert(self, exception):
        alertdict = {
            CertificateError: GNUTLS_A_BAD_CERTIFICATE,
            CertificateAuthorityError: GNUTLS_A_UNKNOWN_CA,
            CertificateSecurityError: GNUTLS_A_INSUFFICIENT_SECURITY,
            CertificateExpiredError: GNUTLS_A_CERTIFICATE_EXPIRED,
            CertificateRevokedError: GNUTLS_A_CERTIFICATE_REVOKED}
        alert = alertdict.get(exception.__class__)
        if alert:
            gnutls_alert_send(self._c_object, GNUTLS_AL_FATAL, alert)

    @method_args(one_of(SHUT_RDWR, SHUT_WR))
    def bye(self, how=SHUT_RDWR):
        gnutls_bye(self._c_object, how)

    def shutdown(self, how=SOCKET_SHUT_RDWR):
        self.socket.shutdown(how)

    def close(self):
        self.socket.close()

    def verify_peer(self):
        status = c_uint()
        gnutls_certificate_verify_peers2(self._c_object, byref(status))
        status = status.value
        if status & GNUTLS_CERT_INVALID:
            raise CertificateError("peer certificate is invalid")
        elif status & GNUTLS_CERT_SIGNER_NOT_FOUND:
            raise CertificateAuthorityError("peer certificate signer not found")
        elif status & GNUTLS_CERT_SIGNER_NOT_CA:
            raise CertificateAuthorityError("peer certificate signer is not a CA")
        elif status & GNUTLS_CERT_INSECURE_ALGORITHM:
            raise CertificateSecurityError("peer certificate uses an insecure algorithm")
        elif status & GNUTLS_CERT_REVOKED:
            raise CertificateRevokedError("peer certificate was revoked")


class ClientSession(Session):
    session_type = GNUTLS_CLIENT

    def __init__(self, socket, context, server_name=None):
        Session.__init__(self, socket, context)
        self._server_name = None
        if server_name is not None:
            self.server_name = server_name

    def _get_server_name(self):
        return self._server_name
    @method_args(str)
    def _set_server_name(self, server_name):
        gnutls_server_name_set(self._c_object, GNUTLS_NAME_DNS, c_char_p(server_name), len(server_name))
        self._server_name = server_name
    server_name = property(_get_server_name, _set_server_name)
    del _get_server_name, _set_server_name


class ServerSession(Session):
    session_type = GNUTLS_SERVER

    def __init__(self, socket, context):
        Session.__init__(self, socket, context)
        gnutls_certificate_server_set_request(self._c_object, CERT_REQUEST)

    @property
    def server_name(self):
        data_length = c_size_t(256)
        data = create_string_buffer(data_length.value)
        hostname_type = c_uint()
        for i in xrange(2**16):
            try:
                gnutls_server_name_get(self._c_object, data, byref(data_length), byref(hostname_type), i)
            except RequestedDataNotAvailable:
                break
            except MemoryError:
                data_length.value += 1 ## one extra byte for the terminating 0
                data = create_string_buffer(data_length.value)
                gnutls_server_name_get(self._c_object, data, byref(data_length), byref(hostname_type), i)
            if hostname_type.value != GNUTLS_NAME_DNS:
                continue
            return data.value
        return None


class ServerSessionFactory(object):

    def __init__(self, socket, context, session_class=ServerSession):
        if not issubclass(session_class, ServerSession):
            raise TypeError, "session_class must be a subclass of ServerSession"
        self.socket = socket
        self.context = context
        self.session_class = session_class

    def __getattr__(self, name):
        ## Generic wrapper for the underlying socket methods and attributes
        return getattr(self.socket, name)

    def bind(self, address):
        self.socket.bind(address)

    def listen(self, backlog):
        self.socket.listen(backlog)

    def accept(self):
        new_sock, address = self.socket.accept()
        session = self.session_class(new_sock, self.context)
        return (session, address)

    def shutdown(self, how=SOCKET_SHUT_RDWR):
        self.socket.shutdown(how)

    def close(self):
        self.socket.close()





"""GNUTLS data validators"""

__all__ = ['function_args', 'method_args', 'none', 'ignore', 'list_of', 'one_of']


# Helper functions (internal use)
#

def isclass(obj):
    return hasattr(obj, '__bases__') or isinstance(obj, type)

# Internal validator classes
#

class Validator(object):
    _registered = []
    def __init__(self, typ):
        self.type = typ
    def check(self, value):
        return False
    @staticmethod
    def can_validate(typ):
        return False
    @classmethod
    def register(cls, validator):
        cls._registered.append(validator)
    @classmethod
    def get(cls, typ):
        for validator in cls._registered:
            if validator.can_validate(typ):
                return validator(typ)
        else:
            return None
    @staticmethod
    def join_names(names):
        if type(names) in (tuple, list):
            if len(names) <= 2:
                return ' or '.join(names)
            else:
                return ' or '.join((', '.join(names[:-1]), names[-1]))
        else:
            return names
    def _type_names(self):
        if isinstance(self.type, tuple):
            return self.join_names([t.__name__.replace('NoneType', 'None') for t in self.type])
        else:
            return self.type.__name__.replace('NoneType', 'None')
    @property
    def name(self):
        name = self._type_names()
        if name.startswith('None'):
            prefix = ''
        elif name[0] in ('a', 'e', 'i', 'o', 'u'):
            prefix = 'an '
        else:
            prefix = 'a '
        return prefix + name

class IgnoringValidator(Validator):
    def __init__(self, typ):
        self.type = none
    def check(self, value):
        return True
    @staticmethod
    def can_validate(obj):
        return obj is ignore

class TypeValidator(Validator):
    def check(self, value):
        return isinstance(value, self.type)
    @staticmethod
    def can_validate(obj):
        return isclass(obj)

class MultiTypeValidator(TypeValidator):
    @staticmethod
    def can_validate(obj):
        return isinstance(obj, tuple) and not filter(lambda x: not isclass(x), obj)

class OneOfValidator(Validator):
    def __init__(self, typ):
        self.type = typ.type
    def check(self, value):
        return value in self.type
    @staticmethod
    def can_validate(obj):
        return isinstance(obj, one_of)
    @property
    def name(self):
        return 'one of %s' % self.join_names(["`%r'" % e for e in self.type])

class ListOfValidator(Validator):
    def __init__(self, typ):
        self.type = typ.type
    def check(self, value):
        return isinstance(value, (tuple, list)) and not filter(lambda x: not isinstance(x, self.type), value)
    @staticmethod
    def can_validate(obj):
        return isinstance(obj, list_of)
    @property
    def name(self):
        return 'a list of %s' % self._type_names()

class ComplexValidator(Validator):
    def __init__(self, typ):
        self.type = [Validator.get(x) for x in typ]
    def check(self, value):
        return bool(sum(t.check(value) for t in self.type))
    @staticmethod
    def can_validate(obj):
        return isinstance(obj, tuple) and not filter(lambda x: Validator.get(x) is None, obj)
    @property
    def name(self):
        return self.join_names([x.name for x in self.type])

Validator.register(IgnoringValidator)
Validator.register(TypeValidator)
Validator.register(MultiTypeValidator)
Validator.register(OneOfValidator)
Validator.register(ListOfValidator)
Validator.register(ComplexValidator)


# Extra types to be used with argument validating decorators
#

none = type(None)

class one_of(object):
    def __init__(self, *args):
        if len(args) < 2:
            raise ValueError("one_of must have at least 2 arguments")
        self.type = args

class list_of(object):
    def __init__(self, *args):
        if filter(lambda x: not isclass(x), args):
            raise TypeError("list_of arguments must be types")
        if len(args) == 1:
            self.type = args[0]
        else:
            self.type = args

ignore = type('ignore', (), {})()


# Helpers for writing well behaved decorators
#

def decorator(func):
    """A syntactic marker with no other effect than improving readability."""
    return func

def preserve_signature(func):
    """Preserve the original function signature and attributes in decorator wrappers."""
    from inspect import getargspec, formatargspec
    from gnutls.constants import GNUTLSConstant
    constants  = [c for c in (getargspec(func)[3] or []) if isinstance(c, GNUTLSConstant)]
    signature  = formatargspec(*getargspec(func))[1:-1]
    parameters = formatargspec(*getargspec(func), **{'formatvalue': lambda value: ""})[1:-1]
    def fix_signature(wrapper):
        if constants:
            ## import the required GNUTLSConstants used as function default arguments
            code = "from gnutls.constants import %s\n" % ', '.join(c.name for c in constants)
            exec code in locals(), locals()
        code = "def %s(%s): return wrapper(%s)\nnew_wrapper = %s\n" % (func.__name__, signature, parameters, func.__name__)
        exec code in locals(), locals()
        new_wrapper.__name__ = func.__name__
        new_wrapper.__doc__ = func.__doc__
        new_wrapper.__module__ = func.__module__
        new_wrapper.__dict__.update(func.__dict__)
        return new_wrapper
    return fix_signature

# Argument validating decorators
#

def _callable_args(*args, **kwargs):
    """Internal function used by argument checking decorators"""
    start = kwargs.get('_start', 0)
    validators = []
    for i, arg in enumerate(args):
        validator = Validator.get(arg)
        if validator is None:
            raise TypeError("unsupported type `%r' at position %d for argument checking decorator" % (arg, i+1))
        validators.append(validator)
    def check_args_decorator(func):
        @preserve_signature(func)
        def check_args(*func_args):
            pos = start
            for validator in validators:
                if not validator.check(func_args[pos]):
                    raise TypeError("argument %d must be %s" % (pos+1-start, validator.name))
                pos += 1
            return func(*func_args)
        return check_args
    return check_args_decorator

@decorator
def method_args(*args):
    """Check class or instance method arguments"""
    return _callable_args(*args, **{'_start': 1})

@decorator
def function_args(*args):
    """Check functions or staticmethod arguments"""
    return _callable_args(*args)




#!/usr/bin/env python

"""Asynchronous client using Twisted with GNUTLS"""

import sys
import os

from twisted.internet.error import ConnectionDone
from twisted.internet.protocol import ClientFactory
from twisted.protocols.basic import LineOnlyReceiver
from twisted.internet import reactor

from gnutls.constants import *
from gnutls.crypto import *
from gnutls.errors import *
from gnutls.interfaces.twisted import TLSContext, X509Credentials

class EchoProtocol(LineOnlyReceiver):

    def connectionMade(self):
        self.sendLine('echo')

    def lineReceived(self, line):
        print 'received: %s' % line
        self.transport.loseConnection()

    def connectionLost(self, reason):
        if reason.type != ConnectionDone:
            print "connection was lost: %s" % reason.value
        reactor.stop()

class EchoFactory(ClientFactory):
    protocol = EchoProtocol

    def clientConnectionFailed(self, connector, err):
        print "connection failed: %s" % err.value
        reactor.stop()


script_path = os.path.realpath(os.path.dirname(sys.argv[0]))
certs_path = os.path.join(script_path, 'certs')

cert = X509Certificate(open(certs_path + '/valid.crt').read())
key = X509PrivateKey(open(certs_path + '/valid.key').read())
ca = X509Certificate(open(certs_path + '/ca.pem').read())
crl = X509CRL(open(certs_path + '/crl.pem').read())
cred = X509Credentials(cert, key, [ca])
cred.verify_peer = True
context = TLSContext(cred, session_parameters="NORMAL:-COMP-ALL:+COMP-DEFLATE:+COMP-NULL")

reactor.connectTLS('localhost', 10000, EchoFactory(), context)
reactor.run()





"""GNUTLS errors"""

__all__ = ['Error', 'GNUTLSError', 'OperationWouldBlock', 'OperationInterrupted',
           'CertificateError', 'CertificateAuthorityError', 'CertificateSecurityError',
           'CertificateExpiredError', 'CertificateRevokedError', 'RequestedDataNotAvailable']

class Error(Exception): pass

class GNUTLSError(Error): pass
class OperationWouldBlock(GNUTLSError): pass
class OperationInterrupted(GNUTLSError): pass

class CertificateError(GNUTLSError): pass
class CertificateAuthorityError(CertificateError): pass
class CertificateSecurityError(CertificateError): pass
class CertificateExpiredError(CertificateError): pass
class CertificateRevokedError(CertificateError): pass

class RequestedDataNotAvailable(GNUTLSError): pass




"""GNUTLS Twisted interface"""

__all__ = ['X509Credentials', 'TLSContext', 'connectTLS', 'listenTLS']

from time import time

from twisted.python import failure
from twisted.internet import main, base, interfaces, abstract, tcp, error

from zope.interface import implementsOnly, implementedBy

from gnutls.connection import ClientSession, ServerSession, ServerSessionFactory
from gnutls.connection import TLSContext, X509Credentials as _X509Credentials
from gnutls.constants import SHUT_RDWR, SHUT_WR
from gnutls.errors import *


class KeepRunning:
    """Return this class from a recurrent function to indicate that it should keep running"""
    pass

class RecurrentCall(object):
    """Execute a function repeatedly at the given interval, until signaled to stop"""
    def __init__(self, period, func, *args, **kwargs):
        from twisted.internet import reactor
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.period = period
        self.now = None
        self.next = None
        self.callid = reactor.callLater(period, self)
    def __call__(self):
        from twisted.internet import reactor
        self.callid = None
        if self.now is None:
            self.now = time()
            self.next = self.now + self.period
        else:
            self.now, self.next = self.next, self.next + self.period
        result = self.func(*self.args, **self.kwargs)
        if result is KeepRunning:
            delay = max(self.next-time(), 0)
            self.callid = reactor.callLater(delay, self)
    def cancel(self):
        if self.callid is not None:
            try:
                self.callid.cancel()
            except ValueError:
                pass
            self.callid = None


class CertificateOK: pass

class X509Credentials(_X509Credentials):
    """A Twisted enhanced X509Credentials"""
    verify_peer = False
    verify_period = None

    def verify_callback(self, peer_cert, preverify_status=None):
        """Verifies the peer certificate and raises an exception if it cannot be accepted"""
        if isinstance(preverify_status, Exception):
            raise preverify_status
        self.check_certificate(peer_cert, cert_name='peer certificate')


class TLSMixin:
    """TLS specific functionality common to both clients and servers"""

    def getPeerCertificate(self):
        return self.socket.peer_certificate

    def doRead(self):
        try:
            return tcp.Connection.doRead(self)
        except (OperationWouldBlock, OperationInterrupted):
            return
        except GNUTLSError, e:
            return e

    def writeSomeData(self, data):
        try:
            return tcp.Connection.writeSomeData(self, data)
        except OperationInterrupted:
            return self.writeSomeData(data)
        except OperationWouldBlock:
            return 0
        except GNUTLSError, e:
            return e

    def _sendCloseReason(self, reason):
        try:
            self.socket.send_alert(reason)
        except OperationInterrupted:
            self._sendCloseReason(reason)

    def _sendCloseAlert(self, how=SHUT_RDWR):
        try:
            self.socket.bye(how)
        except OperationInterrupted:
            self._sendCloseAlert(how)

    def closeTLSSession(self, reason):
        try:
            self._sendCloseReason(reason)
            self._sendCloseAlert(SHUT_RDWR)
        except Exception:
            pass

    def _postLoseConnection(self):
        self.closeTLSSession(self._close_reason)
        return self._close_reason

    def endTLSWrite(self):
        self.stopWriting()
        try:
            self._sendCloseAlert(SHUT_WR)
        except OperationWouldBlock, e:
            if self.socket.interrupted_while_writing:
                self.startWriting()
                return
        except Exception, e:
            return e
        del self.doWrite

    def _closeWriteConnection(self):
        self.doWrite = self.endTLSWrite
        result = self.endTLSWrite()
        if isinstance(result, Exception):
            return result
        return tcp.Connection._closeWriteConnection(self)


class TLSClient(TLSMixin, tcp.Client):
    """Add TLS capabilities to a TCP client"""
    
    implementsOnly(interfaces.ISSLTransport, *[i for i in implementedBy(tcp.Client) if i != interfaces.ITLSTransport])
    
    def __init__(self, host, port, bindAddress, context, connector, reactor=None, server_name=None):
        self.context = context
        self.server_name = server_name
        self.__watchdog = None
        tcp.Client.__init__(self, host, port, bindAddress, connector, reactor)

    def createInternetSocket(self):
        sock = tcp.Client.createInternetSocket(self)
        return ClientSession(sock, self.context, self.server_name)

    def _recurrentVerify(self):
        if not self.connected or self.disconnecting:
            return
        try:
            self.context.credentials.verify_callback(self.socket.peer_certificate)
        except Exception, e:
            self.loseConnection(e)
            return
        else:
            return KeepRunning

    def _verifyPeer(self):
        session = self.socket
        credentials = self.context.credentials
        if not credentials.verify_peer:
            return
        try:
            session.verify_peer()
        except Exception, e:
            preverify_status = e
        else:
            preverify_status = CertificateOK
        
        credentials.verify_callback(session.peer_certificate, preverify_status)
        
        if credentials.verify_period > 0:
            self.__watchdog = RecurrentCall(credentials.verify_period, self._recurrentVerify)

    def doHandshake(self):
        self.stopWriting()
        try:
            self.socket.handshake()
        except (OperationWouldBlock, OperationInterrupted):
            if self.socket.interrupted_while_writing:
                self.startWriting()
            return
        except GNUTLSError, e:
            del self.doRead
            self.failIfNotConnected(err = e)
            return
        
        ## reset any references to the old doRead
        del self.doRead
        self.stopReading()
        
        try:
            self._verifyPeer()
        except GNUTLSError, e:
            self.closeTLSSession(e)
            self.failIfNotConnected(err = e)
            return
        except Exception, e:
            self.closeTLSSession(e)
            self.failIfNotConnected(err = error.getConnectError(str(e)))
            return
        
        ## TLS handshake (including certificate verification) finished succesfully
        tcp.Client._connectDone(self)
        
    def startTLS(self):
        self.doRead = self.doHandshake
        self.startReading()
        self.doHandshake()

    def _connectDone(self):
        self.startTLS()

    def loseConnection(self, reason=failure.Failure(main.CONNECTION_DONE)):
        reason = failure.Failure(reason) # accept python exceptions too
        self._close_reason = reason.value
        abstract.FileDescriptor.loseConnection(self, reason)

    def connectionLost(self, reason):
        if self.__watchdog is not None:
            self.__watchdog.cancel()
            self.__watchdog = None
        tcp.Client.connectionLost(self, reason)


class TLSConnector(base.BaseConnector):
    def __init__(self, host, port, factory, context, timeout, bindAddress, reactor=None, server_name=None):
        self.host = host
        self.port = port
        self.bindAddress = bindAddress
        self.context = context
        self.server_name = server_name
        base.BaseConnector.__init__(self, factory, timeout, reactor)

    def _makeTransport(self):
        return TLSClient(self.host, self.port, self.bindAddress, self.context, self, self.reactor, self.server_name)


class TLSServer(TLSMixin, tcp.Server):
    """Add TLS capabilities to a TCP server"""
    
    implementsOnly(interfaces.ISSLTransport, *[i for i in implementedBy(tcp.Server) if i != interfaces.ITLSTransport])
    
    def __init__(self, sock, protocol, client, server, sessionno, *args, **kw):
        self.__watchdog = None
        self.context = server.context
        tcp.Server.__init__(self, sock, protocol, client, server, sessionno, *args, **kw)
        self.protocol.makeConnection = lambda *args: None
        self.protocol.transport = self ## because we may call connectionLost without connectionMade
        self.startTLS()

    def _recurrentVerify(self):
        if not self.connected or self.disconnecting:
            return
        try:
            self.context.credentials.verify_callback(self.socket.peer_certificate)
        except Exception, e:
            self.loseConnection(e)
            return
        else:
            return KeepRunning

    def _verifyPeer(self):
        session = self.socket
        credentials = self.context.credentials
        if not credentials.verify_peer:
            return
        try:
            session.verify_peer()
        except Exception, e:
            preverify_status = e
        else:
            preverify_status = CertificateOK
        
        credentials.verify_callback(session.peer_certificate, preverify_status)
        
        if credentials.verify_period > 0:
            self.__watchdog = RecurrentCall(credentials.verify_period, self._recurrentVerify)

    def doHandshake(self):
        self.stopWriting()
        try:
            self.socket.handshake()
        except (OperationWouldBlock, OperationInterrupted):
            if self.socket.interrupted_while_writing:
                self.startWriting()
            return
        except GNUTLSError, e:
            del self.doRead
            return e
        
        ## reset any references to the old doRead
        del self.doRead
        self.stopReading()
        self.startReading()
        
        try:
            self._verifyPeer()
        except Exception, e:
            self.loseConnection(e)
            return
        
        ## TLS handshake (including certificate verification) finished succesfully
        
        del self.protocol.makeConnection
        self.protocol.makeConnection(self)

    def startTLS(self):
        self.doRead = self.doHandshake
        self.startReading()

    def loseConnection(self, reason=failure.Failure(main.CONNECTION_DONE)):
        reason = failure.Failure(reason) # accept python exceptions too
        self._close_reason = reason.value
        abstract.FileDescriptor.loseConnection(self, reason)

    def connectionLost(self, reason):
        if self.__watchdog is not None:
            self.__watchdog.cancel()
            self.__watchdog = None
        tcp.Server.connectionLost(self, reason)


class TLSPort(tcp.Port):
    """Add TLS capabilities to a TCP port"""

    transport = TLSServer

    def __init__(self, port, factory, context, backlog=50, interface='', reactor=None, session_class=ServerSession):
        tcp.Port.__init__(self, port, factory, backlog, interface, reactor)
        self.context = context
        self.session_class = session_class

    def createInternetSocket(self):
        sock = tcp.Port.createInternetSocket(self)
        return ServerSessionFactory(sock, self.context, self.session_class)


def connectTLS(reactor, host, port, factory, context, timeout=30, bindAddress=None, server_name=None):
    c = TLSConnector(host, port, factory, context, timeout, bindAddress, reactor, server_name)
    c.connect()
    return c


def listenTLS(reactor, port, factory, context, backlog=50, interface='', session_class=ServerSession):
    p = TLSPort(port, factory, context, backlog, interface, reactor, session_class)
    p.startListening()
    return p

## Add the connectTLS and listenTLS methods to the reactor

import new
from twisted.internet.posixbase import PosixReactorBase

method = new.instancemethod(connectTLS, None, PosixReactorBase)
setattr(PosixReactorBase, 'connectTLS', method)

method = new.instancemethod(listenTLS, None, PosixReactorBase)
setattr(PosixReactorBase, 'listenTLS', method)





GNUTLS_AL_FATAL = 2
GNUTLS_AL_WARNING = 1
GNUTLS_A_ACCESS_DENIED = 49
GNUTLS_A_BAD_CERTIFICATE = 42
GNUTLS_A_BAD_RECORD_MAC = 20
GNUTLS_A_CERTIFICATE_EXPIRED = 45
GNUTLS_A_CERTIFICATE_REVOKED = 44
GNUTLS_A_CERTIFICATE_UNKNOWN = 46
GNUTLS_A_CERTIFICATE_UNOBTAINABLE = 111
GNUTLS_A_CLOSE_NOTIFY = 0
GNUTLS_A_DECODE_ERROR = 50
GNUTLS_A_DECOMPRESSION_FAILURE = 30
GNUTLS_A_DECRYPTION_FAILED = 21
GNUTLS_A_DECRYPT_ERROR = 51
GNUTLS_A_EXPORT_RESTRICTION = 60
GNUTLS_A_HANDSHAKE_FAILURE = 40
GNUTLS_A_ILLEGAL_PARAMETER = 47
GNUTLS_A_INNER_APPLICATION_FAILURE = 208
GNUTLS_A_INNER_APPLICATION_VERIFICATION = 209
GNUTLS_A_INSUFFICIENT_SECURITY = 71
GNUTLS_A_INTERNAL_ERROR = 80
GNUTLS_A_NO_RENEGOTIATION = 100
GNUTLS_A_PROTOCOL_VERSION = 70
GNUTLS_A_RECORD_OVERFLOW = 22
GNUTLS_A_SSL3_NO_CERTIFICATE = 41
GNUTLS_A_UNEXPECTED_MESSAGE = 10
GNUTLS_A_UNKNOWN_CA = 48
GNUTLS_A_UNKNOWN_PSK_IDENTITY = 115
GNUTLS_A_UNRECOGNIZED_NAME = 112
GNUTLS_A_UNSUPPORTED_CERTIFICATE = 43
GNUTLS_A_UNSUPPORTED_EXTENSION = 110
GNUTLS_A_USER_CANCELED = 90
GNUTLS_CERT_IGNORE = 0
GNUTLS_CERT_INSECURE_ALGORITHM = 256
GNUTLS_CERT_INVALID = 2
GNUTLS_CERT_REQUEST = 1
GNUTLS_CERT_REQUIRE = 2
GNUTLS_CERT_REVOKED = 32
GNUTLS_CERT_SIGNER_NOT_CA = 128
GNUTLS_CERT_SIGNER_NOT_FOUND = 64
GNUTLS_CIPHER_3DES_CBC = 3
GNUTLS_CIPHER_AES_128_CBC = 4
GNUTLS_CIPHER_AES_256_CBC = 5
GNUTLS_CIPHER_ARCFOUR_128 = 2
GNUTLS_CIPHER_ARCFOUR_40 = 6
GNUTLS_CIPHER_CAMELLIA_128_CBC = 7
GNUTLS_CIPHER_CAMELLIA_256_CBC = 8
GNUTLS_CIPHER_DES_CBC = 91
GNUTLS_CIPHER_NULL = 1
GNUTLS_CIPHER_RC2_40_CBC = 90
GNUTLS_CIPHER_UNKNOWN = 0
GNUTLS_CLIENT = 2
GNUTLS_COMP_DEFLATE = 2
GNUTLS_COMP_NULL = 1
GNUTLS_COMP_UNKNOWN = 0
GNUTLS_CRD_ANON = 2
GNUTLS_CRD_CERTIFICATE = 1
GNUTLS_CRD_IA = 5
GNUTLS_CRD_PSK = 4
GNUTLS_CRD_SRP = 3
GNUTLS_CRL_REASON_AA_COMPROMISE = 32768 # Variable c_int
GNUTLS_CRL_REASON_AFFILIATION_CHANGED = 16 # Variable c_int
GNUTLS_CRL_REASON_CA_COMPROMISE = 32 # Variable c_int
GNUTLS_CRL_REASON_CERTIFICATE_HOLD = 2 # Variable c_int
GNUTLS_CRL_REASON_CESSATION_OF_OPERATION = 4 # Variable c_int
GNUTLS_CRL_REASON_KEY_COMPROMISE = 64 # Variable c_int
GNUTLS_CRL_REASON_PRIVILEGE_WITHDRAWN = 1 # Variable c_int
GNUTLS_CRL_REASON_SUPERSEEDED = 8 # Variable c_int
GNUTLS_CRL_REASON_UNUSED = 128 # Variable c_int
GNUTLS_CRT_OPENPGP = 2
GNUTLS_CRT_PRINT_FULL = 0
GNUTLS_CRT_PRINT_ONELINE = 1
GNUTLS_CRT_PRINT_UNSIGNED_FULL = 2
GNUTLS_CRT_UNKNOWN = 0
GNUTLS_CRT_X509 = 1
GNUTLS_DIG_MD2 = 5
GNUTLS_DIG_MD5 = 2
GNUTLS_DIG_NULL = 1
GNUTLS_DIG_RMD160 = 4
GNUTLS_DIG_SHA1 = 3
GNUTLS_DIG_SHA224 = 9
GNUTLS_DIG_SHA256 = 6
GNUTLS_DIG_SHA384 = 7
GNUTLS_DIG_SHA512 = 8
GNUTLS_E_AGAIN = -28 # Variable c_int
GNUTLS_E_APPLICATION_ERROR_MAX = -65000 # Variable c_int
GNUTLS_E_APPLICATION_ERROR_MIN = -65500 # Variable c_int
GNUTLS_E_ASN1_DER_ERROR = -69 # Variable c_int
GNUTLS_E_ASN1_DER_OVERFLOW = -77 # Variable c_int
GNUTLS_E_ASN1_ELEMENT_NOT_FOUND = -67 # Variable c_int
GNUTLS_E_ASN1_GENERIC_ERROR = -71 # Variable c_int
GNUTLS_E_ASN1_IDENTIFIER_NOT_FOUND = -68 # Variable c_int
GNUTLS_E_ASN1_SYNTAX_ERROR = -76 # Variable c_int
GNUTLS_E_ASN1_TAG_ERROR = -73 # Variable c_int
GNUTLS_E_ASN1_TAG_IMPLICIT = -74 # Variable c_int
GNUTLS_E_ASN1_TYPE_ANY_ERROR = -75 # Variable c_int
GNUTLS_E_ASN1_VALUE_NOT_FOUND = -70 # Variable c_int
GNUTLS_E_ASN1_VALUE_NOT_VALID = -72 # Variable c_int
GNUTLS_E_BASE64_DECODING_ERROR = -34 # Variable c_int
GNUTLS_E_BASE64_ENCODING_ERROR = -201 # Variable c_int
GNUTLS_E_BASE64_UNEXPECTED_HEADER_ERROR = -207 # Variable c_int
GNUTLS_E_CERTIFICATE_ERROR = -43 # Variable c_int
GNUTLS_E_CERTIFICATE_KEY_MISMATCH = -60 # Variable c_int
GNUTLS_E_COMPRESSION_FAILED = -27 # Variable c_int
GNUTLS_E_CONSTRAINT_ERROR = -101 # Variable c_int
GNUTLS_E_CRYPTO_ALREADY_REGISTERED = -209 # Variable c_int
GNUTLS_E_DB_ERROR = -30 # Variable c_int
GNUTLS_E_DECOMPRESSION_FAILED = -26 # Variable c_int
GNUTLS_E_DECRYPTION_FAILED = -24 # Variable c_int
GNUTLS_E_DH_PRIME_UNACCEPTABLE = -63 # Variable c_int
GNUTLS_E_ENCRYPTION_FAILED = -40 # Variable c_int
GNUTLS_E_ERROR_IN_FINISHED_PACKET = -18 # Variable c_int
GNUTLS_E_EXPIRED = -29 # Variable c_int
GNUTLS_E_FATAL_ALERT_RECEIVED = -12 # Variable c_int
GNUTLS_E_FILE_ERROR = -64 # Variable c_int
GNUTLS_E_GOT_APPLICATION_DATA = -38 # Variable c_int
GNUTLS_E_HANDSHAKE_TOO_LARGE = -210 # Variable c_int
GNUTLS_E_HASH_FAILED = -33 # Variable c_int
GNUTLS_E_IA_VERIFY_FAILED = -104 # Variable c_int
GNUTLS_E_ILLEGAL_SRP_USERNAME = -90 # Variable c_int
GNUTLS_E_INCOMPATIBLE_CRYPTO_LIBRARY = -202 # Variable c_int
GNUTLS_E_INCOMPATIBLE_GCRYPT_LIBRARY = -202 # Variable c_int
GNUTLS_E_INCOMPATIBLE_LIBTASN1_LIBRARY = -203 # Variable c_int
GNUTLS_E_INIT_LIBEXTRA = -82 # Variable c_int
GNUTLS_E_INSUFFICIENT_CREDENTIALS = -32 # Variable c_int
GNUTLS_E_INTERNAL_ERROR = -59 # Variable c_int
GNUTLS_E_INTERRUPTED = -52 # Variable c_int
GNUTLS_E_INVALID_PASSWORD = -99 # Variable c_int
GNUTLS_E_INVALID_REQUEST = -50 # Variable c_int
GNUTLS_E_INVALID_SESSION = -10 # Variable c_int
GNUTLS_E_KEY_USAGE_VIOLATION = -48 # Variable c_int
GNUTLS_E_LARGE_PACKET = -7 # Variable c_int
GNUTLS_E_LIBRARY_VERSION_MISMATCH = -83 # Variable c_int
GNUTLS_E_MAC_VERIFY_FAILED = -100 # Variable c_int
GNUTLS_E_MEMORY_ERROR = -25 # Variable c_int
GNUTLS_E_MPI_PRINT_FAILED = -35 # Variable c_int
GNUTLS_E_MPI_SCAN_FAILED = -23 # Variable c_int
GNUTLS_E_NO_CERTIFICATE_FOUND = -49 # Variable c_int
GNUTLS_E_NO_CIPHER_SUITES = -87 # Variable c_int
GNUTLS_E_NO_COMPRESSION_ALGORITHMS = -86 # Variable c_int
GNUTLS_E_NO_TEMPORARY_DH_PARAMS = -93 # Variable c_int
GNUTLS_E_NO_TEMPORARY_RSA_PARAMS = -84 # Variable c_int
GNUTLS_E_OPENPGP_FINGERPRINT_UNSUPPORTED = -94 # Variable c_int
GNUTLS_E_OPENPGP_GETKEY_FAILED = -88 # Variable c_int
GNUTLS_E_OPENPGP_KEYRING_ERROR = -204 # Variable c_int
GNUTLS_E_OPENPGP_SUBKEY_ERROR = -208 # Variable c_int
GNUTLS_E_OPENPGP_UID_REVOKED = -79 # Variable c_int
GNUTLS_E_PKCS1_WRONG_PAD = -57 # Variable c_int
GNUTLS_E_PK_DECRYPTION_FAILED = -45 # Variable c_int
GNUTLS_E_PK_ENCRYPTION_FAILED = -44 # Variable c_int
GNUTLS_E_PK_SIGN_FAILED = -46 # Variable c_int
GNUTLS_E_PK_SIG_VERIFY_FAILED = -89 # Variable c_int
GNUTLS_E_PULL_ERROR = -54 # Variable c_int
GNUTLS_E_PUSH_ERROR = -53 # Variable c_int
GNUTLS_E_RANDOM_FAILED = -206 # Variable c_int
GNUTLS_E_RECEIVED_ILLEGAL_EXTENSION = -58 # Variable c_int
GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER = -55 # Variable c_int
GNUTLS_E_RECORD_LIMIT_REACHED = -39 # Variable c_int
GNUTLS_E_REHANDSHAKE = -37 # Variable c_int
GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE = -56 # Variable c_int
GNUTLS_E_SHORT_MEMORY_BUFFER = -51 # Variable c_int
GNUTLS_E_SRP_PWD_ERROR = -31 # Variable c_int
GNUTLS_E_SRP_PWD_PARSING_ERROR = -91 # Variable c_int
GNUTLS_E_SUCCESS = 0 # Variable c_int
GNUTLS_E_TOO_MANY_EMPTY_PACKETS = -78 # Variable c_int
GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET = -19 # Variable c_int
GNUTLS_E_UNEXPECTED_PACKET = -15 # Variable c_int
GNUTLS_E_UNEXPECTED_PACKET_LENGTH = -9 # Variable c_int
GNUTLS_E_UNIMPLEMENTED_FEATURE = -1250 # Variable c_int
GNUTLS_E_UNKNOWN_ALGORITHM = -105 # Variable c_int
GNUTLS_E_UNKNOWN_CIPHER_SUITE = -21 # Variable c_int
GNUTLS_E_UNKNOWN_CIPHER_TYPE = -6 # Variable c_int
GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM = -3 # Variable c_int
GNUTLS_E_UNKNOWN_HASH_ALGORITHM = -96 # Variable c_int
GNUTLS_E_UNKNOWN_PKCS_BAG_TYPE = -98 # Variable c_int
GNUTLS_E_UNKNOWN_PKCS_CONTENT_TYPE = -97 # Variable c_int
GNUTLS_E_UNKNOWN_PK_ALGORITHM = -80 # Variable c_int
GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE = -61 # Variable c_int
GNUTLS_E_UNSUPPORTED_VERSION_PACKET = -8 # Variable c_int
GNUTLS_E_UNWANTED_ALGORITHM = -22 # Variable c_int
GNUTLS_E_WARNING_ALERT_RECEIVED = -16 # Variable c_int
GNUTLS_E_WARNING_IA_FPHF_RECEIVED = -103 # Variable c_int
GNUTLS_E_WARNING_IA_IPHF_RECEIVED = -102 # Variable c_int
GNUTLS_E_X509_UNKNOWN_SAN = -62 # Variable c_int
GNUTLS_E_X509_UNSUPPORTED_ATTRIBUTE = -95 # Variable c_int
GNUTLS_E_X509_UNSUPPORTED_CRITICAL_EXTENSION = -47 # Variable c_int
GNUTLS_E_X509_UNSUPPORTED_OID = -205 # Variable c_int
GNUTLS_HANDSHAKE_CERTIFICATE_PKT = 11
GNUTLS_HANDSHAKE_CERTIFICATE_REQUEST = 13
GNUTLS_HANDSHAKE_CERTIFICATE_VERIFY = 15
GNUTLS_HANDSHAKE_CLIENT_HELLO = 1
GNUTLS_HANDSHAKE_CLIENT_KEY_EXCHANGE = 16
GNUTLS_HANDSHAKE_FINISHED = 20
GNUTLS_HANDSHAKE_HELLO_REQUEST = 0
GNUTLS_HANDSHAKE_SERVER_HELLO = 2
GNUTLS_HANDSHAKE_SERVER_HELLO_DONE = 14
GNUTLS_HANDSHAKE_SERVER_KEY_EXCHANGE = 12
GNUTLS_HANDSHAKE_SUPPLEMENTAL = 23
GNUTLS_IA_APPLICATION_PAYLOAD = 0
GNUTLS_IA_FINAL_PHASE_FINISHED = 2
GNUTLS_IA_INTERMEDIATE_PHASE_FINISHED = 1
GNUTLS_KEY_CRL_SIGN = 2 # Variable c_int
GNUTLS_KEY_DATA_ENCIPHERMENT = 16 # Variable c_int
GNUTLS_KEY_DECIPHER_ONLY = 32768 # Variable c_int
GNUTLS_KEY_DIGITAL_SIGNATURE = 128 # Variable c_int
GNUTLS_KEY_ENCIPHER_ONLY = 1 # Variable c_int
GNUTLS_KEY_KEY_AGREEMENT = 8 # Variable c_int
GNUTLS_KEY_KEY_CERT_SIGN = 4 # Variable c_int
GNUTLS_KEY_KEY_ENCIPHERMENT = 32 # Variable c_int
GNUTLS_KEY_NON_REPUDIATION = 64 # Variable c_int
GNUTLS_KP_ANY = '2.5.29.37.0' # Variable STRING
GNUTLS_KP_CODE_SIGNING = '1.3.6.1.5.5.7.3.3' # Variable STRING
GNUTLS_KP_EMAIL_PROTECTION = '1.3.6.1.5.5.7.3.4' # Variable STRING
GNUTLS_KP_OCSP_SIGNING = '1.3.6.1.5.5.7.3.9' # Variable STRING
GNUTLS_KP_TIME_STAMPING = '1.3.6.1.5.5.7.3.8' # Variable STRING
GNUTLS_KP_TLS_WWW_CLIENT = '1.3.6.1.5.5.7.3.2' # Variable STRING
GNUTLS_KP_TLS_WWW_SERVER = '1.3.6.1.5.5.7.3.1' # Variable STRING
GNUTLS_KX_ANON_DH = 4
GNUTLS_KX_DHE_DSS = 2
GNUTLS_KX_DHE_PSK = 10
GNUTLS_KX_DHE_RSA = 3
GNUTLS_KX_PSK = 9
GNUTLS_KX_RSA = 1
GNUTLS_KX_RSA_EXPORT = 6
GNUTLS_KX_SRP = 5
GNUTLS_KX_SRP_DSS = 8
GNUTLS_KX_SRP_RSA = 7
GNUTLS_KX_UNKNOWN = 0
GNUTLS_MAC_MD2 = 5
GNUTLS_MAC_MD5 = 2
GNUTLS_MAC_NULL = 1
GNUTLS_MAC_RMD160 = 4
GNUTLS_MAC_SHA1 = 3
GNUTLS_MAC_SHA256 = 6
GNUTLS_MAC_SHA384 = 7
GNUTLS_MAC_SHA512 = 8
GNUTLS_MAC_UNKNOWN = 0
GNUTLS_MASTER_SIZE = 48 # Variable c_int
GNUTLS_MAX_ALGORITHM_NUM = 16 # Variable c_int
GNUTLS_MAX_SESSION_ID = 32 # Variable c_int
GNUTLS_NAME_DNS = 1
GNUTLS_OID_LDAP_DC = '0.9.2342.19200300.100.1.25' # Variable STRING
GNUTLS_OID_LDAP_UID = '0.9.2342.19200300.100.1.1' # Variable STRING
GNUTLS_OID_PKCS9_EMAIL = '1.2.840.113549.1.9.1' # Variable STRING
GNUTLS_OID_PKIX_COUNTRY_OF_CITIZENSHIP = '1.3.6.1.5.5.7.9.4' # Variable STRING
GNUTLS_OID_PKIX_COUNTRY_OF_RESIDENCE = '1.3.6.1.5.5.7.9.5' # Variable STRING
GNUTLS_OID_PKIX_DATE_OF_BIRTH = '1.3.6.1.5.5.7.9.1' # Variable STRING
GNUTLS_OID_PKIX_GENDER = '1.3.6.1.5.5.7.9.3' # Variable STRING
GNUTLS_OID_PKIX_PLACE_OF_BIRTH = '1.3.6.1.5.5.7.9.2' # Variable STRING
GNUTLS_OID_X520_COMMON_NAME = '2.5.4.3' # Variable STRING
GNUTLS_OID_X520_COUNTRY_NAME = '2.5.4.6' # Variable STRING
GNUTLS_OID_X520_DN_QUALIFIER = '2.5.4.46' # Variable STRING
GNUTLS_OID_X520_GENERATION_QUALIFIER = '2.5.4.44' # Variable STRING
GNUTLS_OID_X520_GIVEN_NAME = '2.5.4.42' # Variable STRING
GNUTLS_OID_X520_INITIALS = '2.5.4.43' # Variable STRING
GNUTLS_OID_X520_LOCALITY_NAME = '2.5.4.7' # Variable STRING
GNUTLS_OID_X520_ORGANIZATIONAL_UNIT_NAME = '2.5.4.11' # Variable STRING
GNUTLS_OID_X520_ORGANIZATION_NAME = '2.5.4.10' # Variable STRING
GNUTLS_OID_X520_PSEUDONYM = '2.5.4.65' # Variable STRING
GNUTLS_OID_X520_STATE_OR_PROVINCE_NAME = '2.5.4.8' # Variable STRING
GNUTLS_OID_X520_SURNAME = '2.5.4.4' # Variable STRING
GNUTLS_OID_X520_TITLE = '2.5.4.12' # Variable STRING
GNUTLS_OPENPGP_CERT = 0
GNUTLS_OPENPGP_CERT_FINGERPRINT = 1
GNUTLS_OPENPGP_FMT_BASE64 = 1
GNUTLS_OPENPGP_FMT_RAW = 0
GNUTLS_PARAMS_DH = 2
GNUTLS_PARAMS_RSA_EXPORT = 1
GNUTLS_PKCS_PLAIN = 1
GNUTLS_PKCS_USE_PBES2_3DES = 16
GNUTLS_PKCS_USE_PKCS12_3DES = 2
GNUTLS_PKCS_USE_PKCS12_ARCFOUR = 4
GNUTLS_PKCS_USE_PKCS12_RC2_40 = 8
GNUTLS_PK_DSA = 2
GNUTLS_PK_RSA = 1
GNUTLS_PK_UNKNOWN = 0
GNUTLS_PSK_KEY_HEX = 1
GNUTLS_PSK_KEY_RAW = 0
GNUTLS_RANDOM_SIZE = 32 # Variable c_int
GNUTLS_SAN_DN = 6
GNUTLS_SAN_DNSNAME = 1
GNUTLS_SAN_IPADDRESS = 4
GNUTLS_SAN_OTHERNAME = 5
GNUTLS_SAN_OTHERNAME_XMPP = 1000
GNUTLS_SAN_RFC822NAME = 2
GNUTLS_SAN_URI = 3
GNUTLS_SERVER = 1
GNUTLS_SHUT_RDWR = 0
GNUTLS_SHUT_WR = 1
GNUTLS_SIGN_DSA_SHA1 = 2
GNUTLS_SIGN_RSA_MD2 = 4
GNUTLS_SIGN_RSA_MD5 = 3
GNUTLS_SIGN_RSA_RMD160 = 5
GNUTLS_SIGN_RSA_SHA1 = 1
GNUTLS_SIGN_RSA_SHA224 = 9
GNUTLS_SIGN_RSA_SHA256 = 6
GNUTLS_SIGN_RSA_SHA384 = 7
GNUTLS_SIGN_RSA_SHA512 = 8
GNUTLS_SIGN_UNKNOWN = 0
GNUTLS_SUPPLEMENTAL_USER_MAPPING_DATA = 0
GNUTLS_SSL3 = 1
GNUTLS_TLS1_0 = 2
GNUTLS_TLS1_1 = 3
GNUTLS_TLS1_2 = 4
GNUTLS_VERIFY_ALLOW_ANY_X509_V1_CA_CRT = 8
GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD2 = 16
GNUTLS_VERIFY_ALLOW_SIGN_RSA_MD5 = 32
GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT = 2
GNUTLS_VERIFY_DISABLE_CA_SIGN = 1
GNUTLS_VERIFY_DO_NOT_ALLOW_SAME = 4
GNUTLS_VERSION_UNKNOWN = 255
GNUTLS_X509_CRT_LIST_IMPORT_FAIL_IF_EXCEED = 1
GNUTLS_X509_FMT_DER = 0
GNUTLS_X509_FMT_PEM = 1

__all__ = sorted(name for name in dir() if name.startswith('GNUTLS_'))




#!/usr/bin/env python

"""Asynchronous server using Twisted with GNUTLS"""

import sys
import os

from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineOnlyReceiver
from twisted.internet.error import CannotListenError, ConnectionDone
from twisted.internet import reactor

from gnutls.constants import *
from gnutls.crypto import *
from gnutls.errors import *
from gnutls.interfaces.twisted import TLSContext, X509Credentials

class EchoProtocol(LineOnlyReceiver):

    def connectionMade(self):
        session = self.transport.socket
        try:
            peer_name = session.peer_certificate.subject
        except AttributeError:
            peer_name = 'Unknown'
        print '\nNew connection from: %s' % peer_name
        print 'Protocol:      %s' % session.protocol
        print 'KX algorithm:  %s' % session.kx_algorithm
        print 'Cipher:        %s' % session.cipher
        print 'MAC algorithm: %s' % session.mac_algorithm
        print 'Compression:   %s' % session.compression

    def lineReceived(self, line):
        if line == 'quit':
            self.transport.loseConnection()
            return
        self.sendLine(line)

    def connectionLost(self, reason):
        if reason.type != ConnectionDone:
            print "Connection was lost: %s" % reason.value

class EchoFactory(Factory):
    protocol = EchoProtocol

script_path = os.path.realpath(os.path.dirname(sys.argv[0]))
certs_path = os.path.join(script_path, 'certs')

cert = X509Certificate(open(certs_path + '/valid.crt').read())
key = X509PrivateKey(open(certs_path + '/valid.key').read())
ca = X509Certificate(open(certs_path + '/ca.pem').read())
crl = X509CRL(open(certs_path + '/crl.pem').read())
cred = X509Credentials(cert, key, [ca], [crl])
cred.verify_peer = True
context = TLSContext(cred, session_parameters="NORMAL:+COMP-DEFLATE")

reactor.listenTLS(10000, EchoFactory(), context)
reactor.run()





"""GNUTLS library errors"""

from gnutls.errors import *
from gnutls.errors import __all__

from gnutls.library.constants import GNUTLS_E_AGAIN, GNUTLS_E_INTERRUPTED, GNUTLS_E_NO_CERTIFICATE_FOUND
from gnutls.library.constants import GNUTLS_E_MEMORY_ERROR, GNUTLS_E_SHORT_MEMORY_BUFFER
from gnutls.library.constants import GNUTLS_E_FATAL_ALERT_RECEIVED, GNUTLS_A_BAD_CERTIFICATE
from gnutls.library.constants import GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE
from gnutls.library.constants import GNUTLS_A_UNKNOWN_CA, GNUTLS_A_INSUFFICIENT_SECURITY
from gnutls.library.constants import GNUTLS_A_CERTIFICATE_EXPIRED, GNUTLS_A_CERTIFICATE_REVOKED
from gnutls.library.functions import gnutls_strerror, gnutls_alert_get

class ErrorMessage(str):
    def __new__(cls, code):
        obj = str.__new__(cls, gnutls_strerror(code))
        obj.code = code
        return obj

# Check functions which return an integer status code (negative codes being errors)
#
class ErrorHandler(object):
    alert_map = {
        GNUTLS_A_BAD_CERTIFICATE       : CertificateError("peer rejected our certificate as invalid"),
        GNUTLS_A_UNKNOWN_CA            : CertificateAuthorityError("peer does not trust our certificate authority"),
        GNUTLS_A_INSUFFICIENT_SECURITY : CertificateSecurityError("peer rejected us on insufficient security"),
        GNUTLS_A_CERTIFICATE_EXPIRED   : CertificateExpiredError("peer rejected our certificate as expired"),
        GNUTLS_A_CERTIFICATE_REVOKED   : CertificateRevokedError("peer rejected our certificate as revoked")
    }

    @classmethod
    def check_status(cls, retcode, function, args):
        if retcode >= 0:
            return retcode
        elif retcode == -1:
            raise GNUTLSError(getattr(function, 'errmsg', None) or ErrorMessage(retcode))
        elif retcode == GNUTLS_E_AGAIN:
            raise OperationWouldBlock(gnutls_strerror(retcode))
        elif retcode == GNUTLS_E_INTERRUPTED:
            raise OperationInterrupted(gnutls_strerror(retcode))
        elif retcode in (GNUTLS_E_MEMORY_ERROR, GNUTLS_E_SHORT_MEMORY_BUFFER):
            raise MemoryError(ErrorMessage(retcode))
        elif retcode == GNUTLS_E_NO_CERTIFICATE_FOUND:
            raise CertificateSecurityError(gnutls_strerror(retcode))
        elif retcode == GNUTLS_E_FATAL_ALERT_RECEIVED:
            exception = cls.alert_map.get(gnutls_alert_get(args[0]))
            raise exception and exception.__class__(*exception.args) or GNUTLSError(ErrorMessage(retcode))
        elif retcode == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE:
            raise RequestedDataNotAvailable(gnutls_strerror(retcode))
        else:
            raise GNUTLSError(ErrorMessage(retcode))


# Attach the error checking function to all functions returning integers
#
from gnutls.library import functions
from ctypes import c_int, c_long

for func in (obj for name, obj in functions.__dict__.iteritems() if name in functions.__all__ and obj.restype in (c_int, c_long)):
    func.errcheck = ErrorHandler.check_status

del c_int, c_long, func, functions




#!/usr/bin/env python

"""Asynchronous client using Twisted with GNUTLS"""

import sys
import os

from twisted.internet import reactor
from eventlet.twistedutil.protocol import GreenClientCreator

from gnutls.constants import *
from gnutls.crypto import *
from gnutls.errors import *
from gnutls.interfaces.twisted import X509Credentials

script_path = os.path.realpath(os.path.dirname(sys.argv[0]))
certs_path = os.path.join(script_path, 'certs')

cert = X509Certificate(open(certs_path + '/valid.crt').read())
key = X509PrivateKey(open(certs_path + '/valid.key').read())
ca = X509Certificate(open(certs_path + '/ca.pem').read())
crl = X509CRL(open(certs_path + '/crl.pem').read())
cred = X509Credentials(cert, key, [ca])
cred.verify_peer = True

try:
    conn = GreenClientCreator(reactor).connectTLS('localhost', 10000, cred)
    conn.write('echo\r\n')
    print "received: %s" % conn.recv().rstrip()
    conn.loseConnection()
except CertificateError, e:
    print e





import sys
from ctypes import *

# Type aliases
#

time_t = c_long
size_t = c_size_t
ssize_t = c_long

gnutls_openpgp_keyid_t = c_ubyte * 8
gnutls_transport_ptr_t = c_void_p
gnutls_x509_dn_t = c_void_p


# Enumerations
#

gnutls_alert_description_t = c_int # enum
gnutls_alert_level_t = c_int # enum
gnutls_certificate_import_flags = c_int # enum
gnutls_certificate_print_formats = c_int # enum
gnutls_certificate_request_t = c_int # enum
gnutls_certificate_status_t = c_int # enum
gnutls_certificate_type_t = c_int # enum
gnutls_certificate_verify_flags = c_int # enum
gnutls_cipher_algorithm_t = c_int # enum
gnutls_close_request_t = c_int # enum
gnutls_compression_method_t = c_int # enum
gnutls_connection_end_t = c_int # enum
gnutls_credentials_type_t = c_int # enum
gnutls_digest_algorithm_t = c_int # enum
gnutls_handshake_description_t = c_int # enum
gnutls_ia_apptype_t = c_int # enum
gnutls_kx_algorithm_t = c_int # enum
gnutls_mac_algorithm_t = c_int # enum
gnutls_openpgp_crt_fmt = c_int # enum
gnutls_openpgp_crt_status_t = c_int # enum
gnutls_params_type_t = c_int # enum
gnutls_pk_algorithm_t = c_int # enum
gnutls_pkcs_encrypt_flags_t = c_int # enum
gnutls_privkey_type_t = c_int # enum
gnutls_protocol_t = c_int # enum
gnutls_psk_key_flags = c_int # enum
gnutls_server_name_type_t = c_int # enum
gnutls_sign_algorithm_t = c_int # enum
gnutls_supplemental_data_format_type_t = c_int # enum
gnutls_x509_crt_fmt_t = c_int # enum
gnutls_x509_subject_alt_name_t = c_int # enum

gnutls_certificate_print_formats_t = gnutls_certificate_print_formats
gnutls_openpgp_crt_fmt_t = gnutls_openpgp_crt_fmt


# Unions, structures and pointers to structure types
#

class gnutls_session_int(Structure):
    _fields_ = []
gnutls_session_t = POINTER(gnutls_session_int)

class gnutls_ia_server_credentials_st(Structure):
    _fields_ = []
gnutls_ia_server_credentials_t = POINTER(gnutls_ia_server_credentials_st)

class gnutls_ia_client_credentials_st(Structure):
    _fields_ = []
gnutls_ia_client_credentials_t = POINTER(gnutls_ia_client_credentials_st)

class gnutls_dh_params_int(Structure):
    _fields_ = []
gnutls_dh_params_t = POINTER(gnutls_dh_params_int)

class gnutls_x509_privkey_int(Structure):
    _fields_ = []
gnutls_x509_privkey_t = POINTER(gnutls_x509_privkey_int)
gnutls_rsa_params_t = POINTER(gnutls_x509_privkey_int)

class params(Union):
    _fields_ = [('dh', gnutls_dh_params_t),
                ('rsa_export', gnutls_rsa_params_t)]

class gnutls_pkcs11_privkey_st(Structure):
    _fields_ = []
gnutls_pkcs11_privkey_t = POINTER(gnutls_pkcs11_privkey_st)

class gnutls_priority_st(Structure):
    _fields_ = []
gnutls_priority_t = POINTER(gnutls_priority_st)

class gnutls_datum_t(Structure):
    _fields_ = [('data', POINTER(c_ubyte)),
                ('size', c_uint)]

class gnutls_params_st(Structure):
    _fields_ = [('type', gnutls_params_type_t),
                ('params', params),
                ('deinit', c_int)]

class gnutls_certificate_credentials_st(Structure):
    _fields_ = []
gnutls_certificate_credentials_t = POINTER(gnutls_certificate_credentials_st)
gnutls_certificate_server_credentials = gnutls_certificate_credentials_t
gnutls_certificate_client_credentials = gnutls_certificate_credentials_t

class gnutls_anon_server_credentials_st(Structure):
    _fields_ = []
gnutls_anon_server_credentials_t = POINTER(gnutls_anon_server_credentials_st)

class gnutls_anon_client_credentials_st(Structure):
    _fields_ = []
gnutls_anon_client_credentials_t = POINTER(gnutls_anon_client_credentials_st)

class gnutls_x509_crl_int(Structure):
    _fields_ = []
gnutls_x509_crl_t = POINTER(gnutls_x509_crl_int)

class gnutls_x509_crt_int(Structure):
    _fields_ = []
gnutls_x509_crt_t = POINTER(gnutls_x509_crt_int)

class gnutls_openpgp_keyring_int(Structure):
    _fields_ = []
gnutls_openpgp_keyring_t = POINTER(gnutls_openpgp_keyring_int)

class gnutls_srp_server_credentials_st(Structure):
    _fields_ = []
gnutls_srp_server_credentials_t = POINTER(gnutls_srp_server_credentials_st)

class gnutls_srp_client_credentials_st(Structure):
    _fields_ = []
gnutls_srp_client_credentials_t = POINTER(gnutls_srp_client_credentials_st)

class gnutls_psk_server_credentials_st(Structure):
    _fields_ = []
gnutls_psk_server_credentials_t = POINTER(gnutls_psk_server_credentials_st)

class gnutls_psk_client_credentials_st(Structure):
    _fields_ = []
gnutls_psk_client_credentials_t = POINTER(gnutls_psk_client_credentials_st)

class gnutls_openpgp_crt_int(Structure):
    _fields_ = []
gnutls_openpgp_crt_t = POINTER(gnutls_openpgp_crt_int)

class gnutls_openpgp_privkey_int(Structure):
    _fields_ = []
gnutls_openpgp_privkey_t = POINTER(gnutls_openpgp_privkey_int)

class cert(Union):
    _fields_ = [('x509', POINTER(gnutls_x509_crt_t)),
                ('pgp', gnutls_openpgp_crt_t)]

class key(Union):
    _fields_ = [('x509', gnutls_x509_privkey_t),
                ('pgp', gnutls_openpgp_privkey_t),
                ('pkcs11', gnutls_pkcs11_privkey_t)]

class gnutls_retr2_st(Structure):
    _fields_ = [('cert_type', gnutls_certificate_type_t),
                ('key_type', gnutls_privkey_type_t),
                ('cert', cert),
                ('ncerts', c_uint),
                ('key', key),
                ('deinit_all', c_uint)]

class gnutls_x509_ava_st(Structure):
    _fields_ = [('oid', gnutls_datum_t),
                ('value', gnutls_datum_t),
                ('value_tag', c_ulong)]

class gnutls_pkcs7_int(Structure):
    _fields_ = []
gnutls_pkcs7_t = POINTER(gnutls_pkcs7_int)

class gnutls_x509_crq_int(Structure):
    _fields_ = []
gnutls_x509_crq_t = POINTER(gnutls_x509_crq_int)


# Function type declarations
#

gnutls_alloc_function = CFUNCTYPE(c_void_p, size_t)
gnutls_calloc_function = CFUNCTYPE(c_void_p, size_t, size_t)
gnutls_certificate_retrieve_function = CFUNCTYPE(c_int, gnutls_session_t, POINTER(gnutls_datum_t), c_int, POINTER(gnutls_pk_algorithm_t), c_int, POINTER(gnutls_retr2_st))
gnutls_db_remove_func = CFUNCTYPE(c_int, c_void_p, gnutls_datum_t)
gnutls_db_retr_func = CFUNCTYPE(gnutls_datum_t, c_void_p, gnutls_datum_t)
gnutls_db_store_func = CFUNCTYPE(c_int, c_void_p, gnutls_datum_t, gnutls_datum_t)
gnutls_free_function = CFUNCTYPE(None, c_void_p)
gnutls_handshake_post_client_hello_func = CFUNCTYPE(c_int, gnutls_session_t)
gnutls_ia_avp_func = CFUNCTYPE(c_int, gnutls_session_t, c_void_p, c_char_p, size_t, POINTER(c_char_p), POINTER(size_t))
gnutls_is_secure_function = CFUNCTYPE(c_int, c_void_p)
gnutls_log_func = CFUNCTYPE(None, c_int, c_char_p)
gnutls_openpgp_recv_key_func = CFUNCTYPE(c_int, gnutls_session_t, POINTER(c_ubyte), c_uint, POINTER(gnutls_datum_t))
gnutls_oprfi_callback_func = CFUNCTYPE(c_int, gnutls_session_t, c_void_p, size_t, POINTER(c_ubyte), POINTER(c_ubyte))
gnutls_params_function = CFUNCTYPE(c_int, gnutls_session_t, gnutls_params_type_t, POINTER(gnutls_params_st))
gnutls_psk_client_credentials_function = CFUNCTYPE(c_int, gnutls_session_t, POINTER(c_char_p), POINTER(gnutls_datum_t))
gnutls_psk_server_credentials_function = CFUNCTYPE(c_int, gnutls_session_t, c_char_p, POINTER(gnutls_datum_t))
gnutls_pull_func = CFUNCTYPE(ssize_t, gnutls_transport_ptr_t, c_void_p, size_t)
gnutls_push_func = CFUNCTYPE(ssize_t, gnutls_transport_ptr_t, c_void_p, size_t)
gnutls_realloc_function = CFUNCTYPE(c_void_p, c_void_p, size_t)
gnutls_sign_func = CFUNCTYPE(c_int, gnutls_session_t, c_void_p, gnutls_certificate_type_t, POINTER(gnutls_datum_t), POINTER(gnutls_datum_t), POINTER(gnutls_datum_t))
gnutls_srp_client_credentials_function = CFUNCTYPE(c_int, gnutls_session_t, POINTER(c_char_p), POINTER(c_char_p))
gnutls_srp_server_credentials_function = CFUNCTYPE(c_int, gnutls_session_t, c_char_p, POINTER(gnutls_datum_t), POINTER(gnutls_datum_t), POINTER(gnutls_datum_t), POINTER(gnutls_datum_t))


__all__ = sorted(name for name in sys.modules[__name__].__dict__ if name.startswith('gnutls_') or name in ('size_t', 'ssize_t', 'time_t', 'cert', 'key', 'params'))





"""GNUTLS constants"""

__all__ = [
    ## Credential types
    'CRED_CERTIFICATE', 'CRED_ANON',

    ## X509 certificate/private key formats
    'X509_FMT_DER', 'X509_FMT_PEM',

    ## Miscellaneous
    'CERT_REQUEST', 'CERT_REQUIRE', 'SHUT_RDWR', 'SHUT_WR'
]

__name_map__ = {
    'PROTO_TLS1_2': 'TLS1_2', 'PROTO_TLS1_1': 'TLS1_1', 'PROTO_TLS1_0': 'TLS1_0',
    'PROTO_SSL3': 'SSL3', 'CRED_CERTIFICATE': 'CRD_CERTIFICATE', 'CRED_ANON': 'CRD_ANON'
}


from gnutls.library import constants

class GNUTLSConstant(int):
    def __new__(cls, name):
        gnutls_name = 'GNUTLS_' + __name_map__.get(name, name)
        instance = int.__new__(cls, getattr(constants, gnutls_name))
        instance.name = name
        return instance
    def __repr__(self):
        return self.name

## Generate all exported constants
code = '\n'.join(["%s = GNUTLSConstant('%s')" % (name, name) for name in __all__])
exec code in locals(), globals()
del code, name

del constants



#!/usr/bin/env python

import os

from distutils.core import setup
from gnutls import __info__ as package_info


def find_packages(toplevel):
    return [directory.replace(os.path.sep, '.') for directory, subdirs, files in os.walk(toplevel) if '__init__.py' in files]


setup(
    name=package_info.__project__,
    version=package_info.__version__,

    description=package_info.__summary__,
    long_description=open('README').read(),
    license=package_info.__license__,
    url=package_info.__webpage__,

    author=package_info.__author__,
    author_email=package_info.__email__,

    platforms=["Platform Independent"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],

    packages=find_packages('gnutls')
)

